//
// Open Service Platform
// Copyright (c) 2012 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

/**
 * @file	FIo_ChannelServiceStub.cpp
 * @brief	This is the implementation file for the  _ChannelServiceStub class.
 *
 */

#include <string>
#include <vector>
#include <iostream>
#include <new>

#include <FBaseSysLog.h>
#include <FBaseRt_EventDispatcher.h>
#include "FIo_IpcServer.h"
#include "FIo_ChannelServiceStub.h"
#include "FIo_ChannelMessages.h"
#include "FIo_ChannelService.h"

using namespace Tizen::Base;
using namespace Tizen::Base::Collection;
using namespace Tizen::Base::Runtime;
using namespace Tizen::Io;

using namespace std;

namespace Tizen { namespace Io
{

_ChannelServiceStub::_ChannelServiceStub(void)
	: __pIpcServer(null)
	, __pChannelService(null)
{

}

_ChannelServiceStub::~_ChannelServiceStub(void)
{
	if (__pIpcServer != null)
	{
		__pIpcServer->Stop();
		delete __pIpcServer;
	}
}

result
_ChannelServiceStub::Construct(void)
{
	SysAssertf(__pIpcServer == null, "Already constructed. Calling Construct() twice or more on a same instance is not allowed for this class.");

	result r = E_SUCCESS;
	_IpcServer* pIpcServer = null;

	pIpcServer = new (std::nothrow) _IpcServer;
	SysTryReturnResult(NID_IO, pIpcServer != null, E_OUT_OF_MEMORY, "The memory is insufficient.");

	r = pIpcServer->Construct("osp.io.ipcserver.channelmanager", *this, false);
	SysTryCatch(NID_IO, !IsFailed(r), , r, "[%s] Failed to create IPC server(ChannelService)", GetErrorMessage(r));

	__pIpcServer = pIpcServer;

	return E_SUCCESS;

CATCH:
	delete pIpcServer;

	return r;
}

void
_ChannelServiceStub::SetChannelService(_ChannelService& service)
{
	__pChannelService = &service;
}

bool
_ChannelServiceStub::OnRegisterChannelMessage(const Tizen::Base::String& appId, int* pResult)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.\n");

	int clientId = __pIpcServer->GetClientId();

	*pResult = __pChannelService->RegisterChannel(appId, clientId);

	return true;
}

bool
_ChannelServiceStub::OnSendRequestMessage(const String& src,
										  const String& dest,
										  const ArrayList& args,
										  int requestId,
										  int* pResult)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.\n");

	*pResult = __pChannelService->SendRequest(src, dest, args, requestId);

	const_cast<ArrayList*>(&args)->RemoveAll(true);

	return true;
}

bool
_ChannelServiceStub::OnSendNullRequestMessage(const String& src,
										  const String& dest,
										  int requestId,
										  int* pResult)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.\n");

	*pResult = __pChannelService->SendNullRequest(src, dest, requestId);

	return true;
}

bool
_ChannelServiceStub::OnSendResponseMessage(const String& src,
										   const String& dest,
										   const ArrayList& args,
										   int requestId,
										   int* pResult)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.\n");

	*pResult = __pChannelService->SendResponse(src, dest, args, requestId);

	const_cast<ArrayList*>(&args)->RemoveAll(true);

	return true;
}

bool
_ChannelServiceStub::OnSendNullResponseMessage(const String& src,
										   const String& dest,
										   int requestId,
										   int* pResult)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.\n");

	*pResult = __pChannelService->SendNullResponse(src, dest, requestId);

	return true;
}

result
_ChannelServiceStub::SendRequest(int clientId, const String& src, const String& dest, const ArrayList& args, int requestId)
{
	SysAssertf(__pIpcServer != null, "Not yet constructed. Construct() should be called before use.\n");

	return __pIpcServer->SendResponse(clientId, new ChannelServiceMsg_sendRequestAsync(src, dest, args, requestId));
}

result
_ChannelServiceStub::SendNullRequest(int clientId, const String& src, const String& dest, int requestId)
{
	SysAssertf(__pIpcServer != null, "Not yet constructed. Construct() should be called before use.\n");

	return __pIpcServer->SendResponse(clientId, new ChannelServiceMsg_sendNullRequestAsync(src, dest, requestId));
}

result
_ChannelServiceStub::SendResponse(int clientId, const String& src, const String& dest, const ArrayList& args, int requestId)
{
	SysAssertf(__pIpcServer != null, "Not yet constructed. Construct() should be called before use.\n");

	return __pIpcServer->SendResponse(clientId, new ChannelServiceMsg_sendResponseAsync(src, dest, args, requestId));
}

result
_ChannelServiceStub::SendNullResponse(int clientId, const String& src, const String& dest, int requestId)
{
	SysAssertf(__pIpcServer != null, "Not yet constructed. Construct() should be called before use.\n");

	return __pIpcServer->SendResponse(clientId, new ChannelServiceMsg_sendNullResponseAsync(src, dest, requestId));
}

void
_ChannelServiceStub::OnIpcRequestReceived(_IpcServer& server, const IPC::Message& message)
{
	IPC_BEGIN_MESSAGE_MAP(_ChannelServiceStub, message)
		IPC_MESSAGE_HANDLER_EX(ChannelServiceMsg_register, &server, OnRegisterChannelMessage)
		IPC_MESSAGE_HANDLER_EX(ChannelServiceMsg_sendRequest, &server, OnSendRequestMessage)
		IPC_MESSAGE_HANDLER_EX(ChannelServiceMsg_sendNullRequest, &server, OnSendNullRequestMessage)
		IPC_MESSAGE_HANDLER_EX(ChannelServiceMsg_sendResponse, &server, OnSendResponseMessage)
		IPC_MESSAGE_HANDLER_EX(ChannelServiceMsg_sendNullResponse, &server, OnSendNullResponseMessage)
	IPC_END_MESSAGE_MAP()
}

void
_ChannelServiceStub::OnIpcServerStarted(const _IpcServer& server)
{

}

void
_ChannelServiceStub::OnIpcServerStopped(const _IpcServer& server)
{

}

void
_ChannelServiceStub::OnIpcClientConnected(const _IpcServer& server, int clientId)
{

}

void
_ChannelServiceStub::OnIpcClientDisconnected(const _IpcServer& server, int clientId)
{
	SysAssertf(__pChannelService != null, "Channel service has not been initialized.");
	SysLog(NID_IO, "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
	__pChannelService->UnregisterChannel(clientId);
	SysLog(NID_IO, ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
}

}}
