// // Open Service Platform
// Copyright (c) 2012 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

/**
 * @file	FIo_MessagePortService.cpp
 * @brief	This is the implementation file for the _MessagePortService class.
 *
 */

#include <unique_ptr.h>

#include <bundle.h>
#include <message-port.h>
#include <pkgmgr-info.h>

#include <FBaseStringComparer.h>
#include <FBaseSysLog.h>
#include <FBase_StringConverter.h>

#include "FIo_MessagePortStub.h"
#include "FIo_MessagePortService.h"

using namespace Tizen::Base;
using namespace Tizen::Base::Collection;
using namespace Tizen::Io;
using namespace Tizen::App;

using namespace std;

static const char LOCAL_APPID[] = "LOCAL_APPID";
static const char LOCAL_PORT[] = "LOCAL_PORT";
static const char TRUSTED_LOCAL[] = "TRUSTED_LOCAL";

static const char REMOTE_APPID[] = "REMOTE_APPID";
static const char REMOTE_PORT[] = "REMOTE_PORT";
static const char TRUSTED_REMOTE[] = "TRUSTED_REMOTE";
static const char TRUSTED_MESSAGE[] = "TRUSTED_MESSAGE";

_MessagePortService::_MessagePortService(void)
	: __pStub(null)
{

}

_MessagePortService::~_MessagePortService(void)
{

}

result
_MessagePortService::Construct(_MessagePortStub& stub)
{
	static _StringHashProvider hashProvider;
	static _StringComparer stringComparer;

	__pPorts = new HashMapT <String, int>();
	__pPorts->Construct(0, 0, hashProvider, stringComparer);

	__pTrustedPorts = new HashMapT <String, int>();
	__pTrustedPorts->Construct(0, 0, hashProvider, stringComparer);

	__pStub = &stub;

	return E_SUCCESS;
}

int
_MessagePortService::RegisterMessagePort(int clientId, const BundleBuffer& buffer)
{
	bundle* b = buffer.b;

	String key = GetKey(buffer);
	String trusted(bundle_get_val(b, TRUSTED_LOCAL));

	SysSecureLog(NID_IO, "Register a message port: [%ls], client = %d", key.GetPointer(), clientId);

	result r = E_SUCCESS;
	int value = -1;
	if (trusted.Equals(L"TRUE", false))
	{
		r = __pTrustedPorts->GetValue(key, value);
		if (r == E_SUCCESS)
		{
			SysSecureTryReturn(NID_IO, value != clientId, MESSAGEPORT_ERROR_IO_ERROR, E_SYSTEM,
					"[E_SYSTEM] The local message port (%ls) has already registered", key.GetPointer());

			SysSecureLog(NID_IO, "Remove garbage values : %ls", key.GetPointer());
			__pTrustedPorts->Remove(key);
		}

		__pTrustedPorts->Add(key, clientId);
	}
	else
	{
		r = __pPorts->GetValue(key, value);
		if (r == E_SUCCESS)
		{
			SysSecureTryReturn(NID_IO, value != clientId, MESSAGEPORT_ERROR_IO_ERROR, E_SYSTEM,
					"[E_SYSTEM] The local message port (%ls) has already registered", key.GetPointer());

			SysSecureLog(NID_IO, "Remove garbage values : %ls", key.GetPointer());
			__pPorts->Remove(key);
		}

		__pPorts->Add(key, clientId);
	}

	return 0;
}

int
_MessagePortService::CheckRemotePort(const BundleBuffer& buffer)
{
	bundle* b = buffer.b;

	String key = GetKey(buffer, false);
	String trusted(bundle_get_val(b, TRUSTED_REMOTE));

	SysSecureLog(NID_IO, "Check a remote message port: [%ls]", key.GetPointer());

	bool out = false;

	if (trusted.Equals(L"TRUE", false))
	{
		__pTrustedPorts->ContainsKey(key, out);
	}
	else
	{
		__pPorts->ContainsKey(key, out);
	}

	SysSecureTryReturn(NID_IO, out == true, MESSAGEPORT_ERROR_MESSAGEPORT_NOT_FOUND, E_OBJ_NOT_FOUND,
			"[E_OBJ_NOT_FOUND] The remote message port (%ls) is not found.", key.GetPointer());

	if (trusted.Equals(L"TRUE", false))
	{
		String localAppId = bundle_get_val(b, LOCAL_APPID);
		String remoteAppId = bundle_get_val(b, REMOTE_APPID);

		// Check the preloaded
		if (!IsPreloaded(localAppId, remoteAppId))
		{
			// Check the certificate
			return CheckCertificate(localAppId, remoteAppId);
		}
	}

	return 0;
}

result
_MessagePortService::UnregisterMessagePort(int clientId)
{
	result r = E_OBJ_NOT_FOUND;

	int value = 0;
	int count = __pPorts->GetCount();

	String key;
	IListT<String>* pKeys = __pPorts->GetKeysN();
	SysTryReturnResult(NID_IO, pKeys != null, E_OUT_OF_MEMORY, "The memory is insufficient.");

	for (int i = 0; i < count; i++)
	{
		pKeys->GetAt(i, key);
		__pPorts->GetValue(key, value);
		if (value == clientId)
		{
			SysSecureLog(NID_IO, "Unregister - key = %ls", key.GetPointer());

			__pPorts->Remove(key);

			r = E_SUCCESS;
		}
	}

	delete pKeys;

	count = __pTrustedPorts->GetCount();
	pKeys = __pTrustedPorts->GetKeysN();
	SysTryReturnResult(NID_IO, pKeys != null, E_OUT_OF_MEMORY, "The memory is insufficient.");

	for (int i = 0; i < count; i++)
	{
		pKeys->GetAt(i, key);
		__pTrustedPorts->GetValue(key, value);
		if (value == clientId)
		{
			SysSecureLog(NID_IO, "Unregister - key = %ls", key.GetPointer());

			__pTrustedPorts->Remove(key);

			r = E_SUCCESS;
		}
	}

	delete pKeys;

	return r;
}

int
_MessagePortService::SendMessage(const BundleBuffer& metadata, const BundleBuffer& buffer)
{
	result r = E_SUCCESS;
	int clientId = 0;

	bundle* b = metadata.b;

	String key = GetKey(metadata, false);
	SysSecureLog(NID_IO, "Sends a message to a remote message port [%ls]", key.GetPointer());

	String trustedMessage(bundle_get_val(b, TRUSTED_MESSAGE));
	if (trustedMessage.Equals(L"TRUE", false))
	{
		r = __pTrustedPorts->GetValue(key, clientId);
	}
	else
	{
		r = __pPorts->GetValue(key, clientId);
	}

	SysSecureTryReturn(NID_IO, r == E_SUCCESS, MESSAGEPORT_ERROR_MESSAGEPORT_NOT_FOUND, E_OBJ_NOT_FOUND,
				"[E_OBJ_NOT_FOUND] The destination message port (%ls) is not found.", key.GetPointer());

	if (trustedMessage.Equals(L"TRUE", false))
	{
		String localAppId = bundle_get_val(b, LOCAL_APPID);
		String remoteAppId = bundle_get_val(b, REMOTE_APPID);

		// Check the preloaded
		if (!IsPreloaded(localAppId, remoteAppId))
		{
			// Check the certificate
			int ret = CheckCertificate(localAppId, remoteAppId);
			if (ret < 0)
			{
				return ret;
			}
		}
	}

	r = __pStub->SendMessage(clientId, metadata, buffer);
	SysTryReturn(NID_IO, r == E_SUCCESS, MESSAGEPORT_ERROR_IO_ERROR, E_SYSTEM,
				"[E_SYSTEM] Failed to send a message.");

	return E_SUCCESS;
}

String 
_MessagePortService::GetKey(const BundleBuffer& buffer, bool local) const
{
	const char* pAppId = null; 
	const char* pPortName = null; 

	bundle* b = buffer.b;

	if (local)
	{
		pAppId = bundle_get_val(b, LOCAL_APPID);
		pPortName = bundle_get_val(b, LOCAL_PORT);
	}
	else
	{
		pAppId = bundle_get_val(b, REMOTE_APPID);
		pPortName = bundle_get_val(b, REMOTE_PORT);
	}

	String key(pAppId);
	key.Append(L':');
	key.Append(pPortName);

	return key;
}

bool
_MessagePortService::IsPreloaded(const String& localAppId, const String& remoteAppId) const
{
	bool preload_local = false;
	bool preload_remote = false;

	unique_ptr<char[]> pLocalAppId(_StringConverter::CopyToCharArrayN(localAppId));
	unique_ptr<char[]> pRemoteAppId(_StringConverter::CopyToCharArrayN(remoteAppId));

	pkgmgrinfo_appinfo_h handle = null;
	int ret = pkgmgrinfo_appinfo_get_appinfo(pLocalAppId.get(), &handle);
	if (ret != PMINFO_R_OK)
	{
		SysLog(NID_IO, "Failed to get the appinfo. %d", ret);
		return false;
	}
	ret = pkgmgrinfo_appinfo_is_preload(handle, &preload_local);
	if (ret != PMINFO_R_OK)
	{
		SysLog(NID_IO, "Failed to check the preloaded application. %d", ret);
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return false;
	}

	ret = pkgmgrinfo_appinfo_get_appinfo(pRemoteAppId.get(), &handle);
	if (ret != PMINFO_R_OK)
	{
		SysLog(NID_IO, "Failed to get the appinfo. %d", ret);
		return false;
	}
	ret = pkgmgrinfo_appinfo_is_preload(handle, &preload_remote);
	if (ret != PMINFO_R_OK)
	{
		SysLog(NID_IO, "Failed to check the preloaded application. %d", ret);
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return false;
	}

	if (preload_local && preload_remote)
	{
		return true;
	}

	return false;
}

int
_MessagePortService::CheckCertificate(const String& localAppId, const String& remoteAppId) const
{
	unique_ptr<char[]> pLocalAppId(_StringConverter::CopyToCharArrayN(localAppId));
	unique_ptr<char[]> pRemoteAppId(_StringConverter::CopyToCharArrayN(remoteAppId));

	pkgmgrinfo_cert_compare_result_type_e res;
	int ret = pkgmgrinfo_pkginfo_compare_app_cert_info(pLocalAppId.get(), pRemoteAppId.get(), &res);
	SysTryReturn(NID_IO, ret == 0, MESSAGEPORT_ERROR_IO_ERROR, E_SYSTEM,
				"[E_SYSTEM] Failed to check the certificate: %d", ret);

	SysTryReturn(NID_IO, res == PMINFO_CERT_COMPARE_MATCH, MESSAGEPORT_ERROR_CERTIFICATE_NOT_MATCH, E_CERTIFICATE_VERIFICATION_FAILED,
				"[E_CERTIFICATE_VERIFICATION_FAILED] Both applications are not signed with the same certificate: %d", res);

	return 0;
}

