//
// Open Service Platform
// Copyright (c) 2012 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

/**
 * @file	FIo_MessagePortStub.cpp
 * @brief	This is the implementation file for the  _MessagePortStub class.
 *
 */

#include <message-port-messages.h>

#include <bundle.h>

#include <FBaseSysLog.h>

#include "FIo_MessagePortStub.h"
#include "FIo_MessagePortService.h"

using namespace Tizen::Base;
using namespace Tizen::Base::Collection;
using namespace Tizen::Base::Runtime;
using namespace Tizen::Io;
using namespace Tizen::App;

_MessagePortStub::_MessagePortStub(void)
	: __pIpcServer(null)
	, __pService(null)
{

}

_MessagePortStub::~_MessagePortStub(void)
{
	if (__pIpcServer != null)
	{
		__pIpcServer->Stop();
		delete __pIpcServer;
	}
}

result
_MessagePortStub::Construct(void)
{
	SysAssertf(__pIpcServer == null, "Already constructed. Calling Construct() twice or more on a same instance is not allowed for this class.");

	result r = E_SUCCESS;
	IpcServer* pIpcServer = new (std::nothrow) IpcServer;
	SysTryReturnResult(NID_IO, pIpcServer != null, E_OUT_OF_MEMORY, "Not enough memory.");

	r = pIpcServer->Construct("message-port-server", *this, false);
	SysTryCatch(NID_IO, !IsFailed(r), , r, "[%s] Failed to create IPC server(MessagePort)", GetErrorMessage(r));

	__pIpcServer = pIpcServer;

	return E_SUCCESS;

CATCH:
	delete pIpcServer;

	return r;
}

void
_MessagePortStub::SetMessagePortService(_MessagePortService& service)
{
	__pService = &service;
}

bool
_MessagePortStub::OnRegisterMessagePort(const BundleBuffer& buffer, int* pResult)
{
	SysAssertf(__pService != null, "Message port service has not been initialized.\n");

	int clientId = __pIpcServer->GetClientId();

	*pResult = __pService->RegisterMessagePort(clientId, buffer);

	bundle_free(buffer.b);

	return true;
}

bool
_MessagePortStub::OnCheckRemotePort(const BundleBuffer& buffer, int* pResult)
{
	SysAssertf(__pService != null, "Message port service has not been initialized.\n");

	*pResult = __pService->CheckRemotePort(buffer);

	bundle_free(buffer.b);

	return true;
}

bool
_MessagePortStub::OnSendMessage(const BundleBuffer& metadata, const BundleBuffer& buffer, int* pResult)
{
	SysAssertf(__pService != null, "MessagePort service has not been initialized.\n");

	*pResult = __pService->SendMessage(metadata, buffer);

	bundle_free(metadata.b);
	bundle_free(buffer.b);

	return true;
}

result
_MessagePortStub::SendMessage(int clientId, const BundleBuffer& metadata, const BundleBuffer& buffer)
{
	SysAssertf(__pService != null, "MessagePort service has not been initialized.\n");

	result r = __pIpcServer->SendResponse(clientId, new MessagePort_sendMessageAsync(metadata, buffer));

	SysTryReturnResult(NID_IO, r == E_SUCCESS, r, "Failed to send a response message.");

	return E_SUCCESS;
}

void
_MessagePortStub::OnIpcRequestReceived(IpcServer& server, const IPC::Message& message)
{
	IPC_BEGIN_MESSAGE_MAP(_MessagePortStub, message)
		IPC_MESSAGE_HANDLER_EX(MessagePort_registerPort, &server, OnRegisterMessagePort)
		IPC_MESSAGE_HANDLER_EX(MessagePort_checkRemotePort, &server, OnCheckRemotePort)
		IPC_MESSAGE_HANDLER_EX(MessagePort_sendMessage, &server, OnSendMessage)
	IPC_END_MESSAGE_MAP()
}

void
_MessagePortStub::OnIpcServerStarted(const IpcServer& server)
{

}

void
_MessagePortStub::OnIpcServerStopped(const IpcServer& server)
{

}

void
_MessagePortStub::OnIpcClientConnected(const IpcServer& server, int clientId)
{
	SysLog(NID_IO, "Ipc connected");
}

void
_MessagePortStub::OnIpcClientDisconnected(const IpcServer& server, int clientId)
{
	SysAssertf(__pService != null, "MessagePort service has not been initialized.");
	SysLog(NID_IO, "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
	SysLog(NID_IO, "Unregister - client =  %d", clientId);
	SysLog(NID_IO, "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");

	__pService->UnregisterMessagePort(clientId);
}

