//
// Open Service Platform
// Copyright (c) 2012 Samsung Electronics Co., Ltd.
//
// Licensed under the Apache License, Version 2.0 (the License);
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

/**
 * @file  FIo_IpcServer.cpp
 * @brief This is the implementation file for the IpcServer class.
 *
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cerrno>
#include <iostream>
#include <new>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>

#include <sys/smack.h>

#include <aul/aul.h>

#include <ipc/ipc_message.h>

#include <FBaseRtMutex.h>
#include <FBaseSysLog.h>
#include <FBase_StringConverter.h>
#include <FBaseRt_EventDispatcher.h>

#include "IIpcServerEventListener.h"
#include "IpcServer.h"

using namespace std;
using namespace IPC;

using namespace Tizen::Base;
using namespace Tizen::Base::Runtime;
using namespace Tizen::App;

IpcServer::_ChannelInfo::_ChannelInfo(void)
	: pClientInfo(null)
	, pGIOChannel(null)
	, pGSource(null)
	, destroySource(true)
{

}

IpcServer::_ChannelInfo::~_ChannelInfo(void)
{
	if (pGIOChannel != null)
	{
		g_io_channel_unref(pGIOChannel);
	}

	if (pGSource != null)
	{
		if (destroySource)
		{
			g_source_destroy(pGSource);
		}

		g_source_unref(pGSource);
	}
}

IpcServer::_ClientInfo::_ClientInfo(void)
	: clientId(-1)
	, pIpcServer(null)
	, pReverseChannel(null)
{

}

IpcServer::_ClientInfo::~_ClientInfo(void)
{
	if (pReverseChannel != null)
	{
		g_io_channel_unref(pReverseChannel);
	}

	channels.clear();
}

IpcServer::IpcServer(void)
	: __runOnCallerThread(false)
	, __pEventDispatcher(null)
	, __pListener(null)
	, __handlerThread(0)
	, __pHandlerGMainContext(null)
	, __pHandlerGMainLoop(null)
	, __pConnectGSource(null)
	, __pCurrentChannel(null)
	, __pCurrentClientInfo(null)
{
	__messageBuffer[0] = '\0';
}


IpcServer::~IpcServer(void)
{
	if (__pConnectGSource != null)
	{
		g_source_destroy(__pConnectGSource);
		g_source_unref(__pConnectGSource);
		__pConnectGSource = null;
	}

	if (!__runOnCallerThread)
	{
		if (__pEventDispatcher)
		{
			delete __pEventDispatcher;
			__pEventDispatcher = null;
		}

		if (__pHandlerGMainLoop)
		{
			g_main_loop_unref(__pHandlerGMainLoop);
			__pHandlerGMainLoop = null;
		}

		if (__pHandlerGMainContext)
		{
			g_main_context_unref(__pHandlerGMainContext);
			__pHandlerGMainContext = null;
		}
	}

	// clean-up clients
	__clients.clear();
}

result
IpcServer::Construct(const String& name, const IIpcServerEventListener& listener, bool runOnCallerThread)
{
	result r = E_SUCCESS;

	GIOChannel* pGIOChannel = null;
	GSource* pGSource = null;

	struct sockaddr_un serverAddress;
	int serverSocket = -1;
	int serverLen = 0;
	int ret = 0;
	std::string socketName;
	char* pName = null;
	size_t socketNameLength = 0;

	__name = name;
	__pListener = const_cast <IIpcServerEventListener*>(&listener);
	__runOnCallerThread = runOnCallerThread;

	pName = _StringConverter::CopyToCharArrayN(name);
	SysTryReturnResult(NID_IO, pName != null, E_OUT_OF_MEMORY, "Not enough memory");

	socketName.append("/var/run/osp/");
	socketName.append(pName);

	delete[] pName;

	socketNameLength = socketName.size() + 1;
	SysTryReturnResult(NID_IO, socketNameLength < 108, E_INVALID_ARG, "Server name is too long");

	if (!__runOnCallerThread)
	{
		__pHandlerGMainContext = g_main_context_new();
		__pHandlerGMainLoop = g_main_loop_new(__pHandlerGMainContext, FALSE);

	}
	else
	{
		__pHandlerGMainContext = g_main_context_get_thread_default();
		if (__pHandlerGMainContext == null) // is global?
		{
			__pHandlerGMainContext = g_main_context_default();
			if (__pHandlerGMainContext == null)
			{
				return E_SYSTEM;
			}
		}
	}

	ret = unlink(socketName.c_str());
	SysTryLog(NID_IO, ret == 0, "Unlink a socket has failed.. but it is not a problem.");

	serverSocket = socket(AF_UNIX, SOCK_STREAM, 0);
	SysTryReturnResult(NID_IO, serverSocket != -1, E_SYSTEM, "Failed to create a socket.");

	// SMACK (Add a * label to socket)
	if(smack_fsetlabel(serverSocket, "@", SMACK_LABEL_IPOUT) != 0)
	{
		SysTryCatch(NID_IO, errno == EOPNOTSUPP, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] SMACK labeling failed");
		SysLog(NID_IO, "Kernel doesn't have Smack.");
	}

	if(smack_fsetlabel(serverSocket, "*", SMACK_LABEL_IPIN) != 0)
	{
		SysTryCatch(NID_IO, errno == EOPNOTSUPP, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] SMACK labeling failed");
		SysLog(NID_IO, "Kernel doesn't have Smack.");
	}

	bzero(&serverAddress, sizeof(serverAddress));
	serverAddress.sun_family = AF_UNIX;
	strncpy(serverAddress.sun_path, socketName.c_str(), socketNameLength);
	serverLen = sizeof(serverAddress);

	ret = bind(serverSocket, (const struct sockaddr*) &serverAddress, serverLen);
	SysTryCatch(NID_IO, ret != -1, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] Failed to bind a socket(%d, %s): %s", serverSocket,
			   socketName.c_str(), strerror(errno));

	ret = chmod(socketName.c_str(), 0666);
	SysTryCatch(NID_IO, ret == 0, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] Failed to change permission of a socket(%d, %s): %s", serverSocket,
			   socketName.c_str(), strerror(errno));

	ret = listen(serverSocket, 128);
	SysTryCatch(NID_IO, ret == 0, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] Failed to listen a socket(%d, %s): %s", serverSocket,
			   socketName.c_str(), strerror(errno));

	pGIOChannel = g_io_channel_unix_new(serverSocket);
	SysTryCatch(NID_IO, pGIOChannel != null, , E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] Not enough memory.");

	// socket will be closed when pGIOChannel is deleted.
	g_io_channel_set_close_on_unref(pGIOChannel, TRUE);
	serverSocket = -1;

	pGSource = g_io_create_watch(pGIOChannel, (GIOCondition) (G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP));
	SysTryCatch(NID_IO, pGSource != null, r = E_OUT_OF_MEMORY, E_OUT_OF_MEMORY,
			   "[E_OUT_OF_MEMORY] Failed to create watch for socket.");

	// channel will be delete when pGSource is deleted.
	g_io_channel_unref(pGIOChannel);
	pGIOChannel = null;

	g_source_set_callback(pGSource, (GSourceFunc) OnConnectionRequest, this, NULL);
	g_source_attach(pGSource, __pHandlerGMainContext);

	if (__runOnCallerThread)
	{
		__pListener->OnIpcServerStarted(*this);
	}
	else
	{
		ret = pthread_create(&__handlerThread, null, &ThreadProc, (void*) this);
		SysTryCatch(NID_IO, ret == 0, r = E_SYSTEM, E_SYSTEM, "[E_SYSTEM] Failed to an IPC thread");
	}

	__pConnectGSource = pGSource;

	return E_SUCCESS;

CATCH:
	if (pGIOChannel != null)
	{
		g_io_channel_unref(pGIOChannel);
	}

	if (serverSocket != -1)
	{
		close(serverSocket);
	}

	if (runOnCallerThread && __pHandlerGMainContext)
	{
		g_main_context_unref(__pHandlerGMainContext);
		__pHandlerGMainContext = null;
	}

	return r;
}

struct HelloMessage
{
	int reverse;  // if the connection is for reverse message
};

gboolean
IpcServer::OnConnectionRequest(GIOChannel* source, GIOCondition condition, gpointer data)
{
	IpcServer* pIpcServer = (IpcServer*) data;
	GError* pGError = null;
	HelloMessage helloMessage;
	_ClientInfo* pClientInfo = null;
	_ChannelInfo* pChannelInfo = null;
	GSource* pGSource = null;
	GIOChannel* pChannel = null;
	ssize_t readBytes = 0;
	int ret = 0;

	int server = -1;
	int client = -1;
	struct sockaddr_un clientAddress;
	socklen_t clientLen = sizeof(clientAddress);

	struct ucred cr;
	socklen_t ucredLen = sizeof(cr);

	SysAssertf(pIpcServer != null, "Not yet constructed. Construct() should be called before use.\n");
	SysAssertf(pIpcServer->__pListener != null, "Listener is null.\n");

	server = g_io_channel_unix_get_fd(source);

	client = accept(server, (struct sockaddr*) &clientAddress, &clientLen);
	SysTryCatch(NID_IO, client != -1, , E_SYSTEM, "[E_SYSTEM] Accept failed.");

	readBytes = read(client, &helloMessage, sizeof(helloMessage));
	SysTryCatch(NID_IO, readBytes >= 0, , E_SYSTEM, "[E_SYSTEM] Failed to receive hello message (%d, %s).",
			errno, strerror(errno));

	pChannel = g_io_channel_unix_new(client);
	SysTryCatch(NID_IO, pChannel != null, , E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] Not enough memory.");

	g_io_channel_set_encoding(pChannel, NULL, &pGError);
	g_io_channel_set_flags(pChannel, G_IO_FLAG_NONBLOCK, &pGError);

	g_io_channel_set_close_on_unref(pChannel, TRUE);

	ret = getsockopt(client, SOL_SOCKET, SO_PEERCRED, &cr, &ucredLen);
	SysTryCatch(NID_IO, ret >= 0, , E_SYSTEM, "[E_SYSTEM] Failed to get peercred information: %s", strerror(errno));

	client = -1;

	pClientInfo = pIpcServer->__clients[cr.pid];
	if (pClientInfo == null) // first connection request from this client
	{
		pClientInfo = new (std::nothrow) _ClientInfo;
		SysTryCatch(NID_IO, pClientInfo != null, , E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] Not enough memory.");

		pClientInfo->pIpcServer = pIpcServer;
		pClientInfo->clientId = cr.pid;

		char buffer[256] = {0, };
		ret = aul_app_get_appid_bypid(cr.pid, buffer, sizeof(buffer));
		SysTryCatch(NID_IO, ret == AUL_R_OK, delete pClientInfo, E_SYSTEM, "[E_SYSTEM] Failed to get the application ID of pid: %d", ret);

		pClientInfo->appId = buffer;

		pClientInfo->pReverseChannel = null;

		pIpcServer->__clients[cr.pid] = pClientInfo;
		pIpcServer->__pCurrentClientInfo = pClientInfo;
		pIpcServer->__pListener->OnIpcClientConnected(*pIpcServer, cr.pid);
		pIpcServer->__pCurrentClientInfo = null;
	}

	if (helloMessage.reverse != 0)
	{
		pClientInfo->pReverseChannel = pChannel;
	}
	else
	{
		pChannelInfo = new (std::nothrow) _ChannelInfo;
		SysTryCatch(NID_IO, pChannelInfo != null, , E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] Not enough memory.");

		pGSource = g_io_create_watch(pChannel, (GIOCondition) (G_IO_IN | G_IO_ERR | G_IO_NVAL | G_IO_HUP));
		g_source_set_callback(pGSource, (GSourceFunc) OnReadMessage, pChannelInfo, NULL);
		g_source_attach(pGSource, pIpcServer->__pHandlerGMainContext);

		pChannelInfo->pClientInfo = pClientInfo;
		pChannelInfo->pGIOChannel = pChannel;
		pChannelInfo->pGSource = pGSource;

		pClientInfo->channels.push_back(pChannelInfo);
	}

	return true;

CATCH:
	if (pChannel != null)
	{
		g_io_channel_unref(pChannel);
	}

	if (client != -1)
	{
		close(client);
	}

	return true;
}

int
IpcServer::GetClientId(void) const
{
	if (__pCurrentClientInfo)
	{
		return __pCurrentClientInfo->clientId;
	}

	return -1;
}

AppId
IpcServer::GetClientApplicationId(void) const
{
	static String nullString;

	if (__pCurrentClientInfo)
	{
		return __pCurrentClientInfo->appId;
	}

	return nullString;
}

gboolean
IpcServer::HandleReceivedMessage(GIOChannel* source, GIOCondition condition, gpointer data)
{
	GError* pGError = NULL;
	GIOStatus status;
	IPC::Message* pMessage = NULL;
	_ChannelInfo* pChannelInfo = (_ChannelInfo*) data;
	_ClientInfo* pClientInfo = pChannelInfo->pClientInfo;

	if (condition & G_IO_HUP)
	{
		SysLog(NID_IO, "Connection closed");
		int clientId = pClientInfo->clientId;

		g_io_channel_shutdown(source, FALSE, &pGError);

		for (unsigned int i = 0; i < pClientInfo->channels.size(); i++)
		{
			if (pChannelInfo == pClientInfo->channels[i])
			{
				pClientInfo->channels.erase(pClientInfo->channels.begin() + i);

				// Do not destroy a source in a dispatch callback
				// because main loop will do it if the callback return FALSE.
				pChannelInfo->destroySource = false;
				delete pChannelInfo;

				break;
			}
		}

		if (pClientInfo->channels.size() == 0)
		{
			SysLog(NID_IO, "All connections of client(%d) are closed. delete client info", clientId);

			__pListener->OnIpcClientDisconnected(*this, clientId);

			__clients[clientId] = null;

			delete pClientInfo;
		}

		return FALSE;
	}
	else if (condition & G_IO_IN)
	{
		gsize readSize = 0;
		const char* pStart = NULL;
		const char* pEnd = NULL;
		const char* pEndOfMessage = NULL;

		while (true)
		{
			pGError = null;
			status = g_io_channel_read_chars(source, (char*) __messageBuffer, __MAX_MESSAGE_BUFFER_SIZE, &readSize, &pGError);
			if (status != G_IO_STATUS_NORMAL)
			{
				if (status == G_IO_STATUS_EOF || status == G_IO_STATUS_ERROR)
				{
					if (status == G_IO_STATUS_EOF)
					{
						SysLog(NID_IO, "G_IO_STATUS_EOF, the connection is closed.");
					}
					else
					{
						SysLog(NID_IO, "G_IO_STATUS_ERROR, the connection is closed. ");
					}

					pGError = null;
					g_io_channel_shutdown(source, FALSE, &pGError);

					int clientId = pClientInfo->clientId;

					for (unsigned int i = 0; i < pClientInfo->channels.size(); i++)
					{
						if (pChannelInfo == pClientInfo->channels[i])
						{
							pClientInfo->channels.erase(pClientInfo->channels.begin() + i);

							pChannelInfo->destroySource = false;
							delete pChannelInfo;
							break;
						}
					}

					if (pClientInfo->channels.size() == 0)
					{
						SysLog(NID_IO, "All connections of client(%d) are closed normally by the client.", clientId);

						if (__pListener)
						{
							__pListener->OnIpcClientDisconnected(*this, clientId);
						}

						__clients[clientId] = null;

						delete pClientInfo;
					}

					return FALSE;
				}
			}

			if (readSize == 0)
			{
				break;
			}

			if (__pending.empty())
			{
				pStart = __messageBuffer;
				pEnd = pStart + readSize;
			}
			else
			{
				__pending.append(__messageBuffer, readSize);
				pStart = __pending.data();
				pEnd = pStart + __pending.size();
			}

			while (true)
			{
				pEndOfMessage = IPC::Message::FindNext(pStart, pEnd);
				if (pEndOfMessage == NULL)
				{
					__pending.assign(pStart, pEnd - pStart);
					break;
				}

				pMessage = new (std::nothrow) IPC::Message(pStart, pEndOfMessage - pStart);
				SysTryReturn(NID_IO, pMessage != null, FALSE, E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] The memory is insufficient.");

				__pCurrentChannel = source;

				if (__pListener)
				{
					__pListener->OnIpcRequestReceived(*this, *pMessage);
				}

				delete pMessage;

				__pCurrentChannel = NULL;

				pStart = pEndOfMessage;
			}
		}
	}
	else
	{
		// empty statement
	}

	return TRUE;
}

gboolean
IpcServer::OnReadMessage(GIOChannel* source, GIOCondition condition, gpointer data)
{
	gboolean ret = FALSE;
	_ChannelInfo* pChannelInfo = (_ChannelInfo*) data;
	_ClientInfo* pClientInfo = pChannelInfo->pClientInfo;
	IpcServer* pIpcServer = (IpcServer*) pClientInfo->pIpcServer;

	pIpcServer->__pCurrentClientInfo = pClientInfo;
	ret = pIpcServer->HandleReceivedMessage(source, condition, data);
	pIpcServer->__pCurrentClientInfo = null;

	return ret;
}

void*
IpcServer::ThreadProc(void* pParam)
{
	IpcServer* pIpcServer = (IpcServer*) pParam;
	if (pIpcServer != NULL)
	{
		pIpcServer->Run(NULL);
	}

	return NULL;
}

void
IpcServer::Run(void* pParam)
{
	result r = E_SUCCESS;

	if (__pListener == null)
	{
		return;
	}

	__pEventDispatcher = new (std::nothrow) _EventDispatcher;
	SysTryReturnVoidResult(NID_IO, __pEventDispatcher != null, E_OUT_OF_MEMORY, "[E_OUT_OF_MEMORY] The memory is insufficient.");

	r = __pEventDispatcher->Construct(__pHandlerGMainContext);
	if (IsFailed(r))
	{
		delete __pEventDispatcher;
		__pEventDispatcher = null;
	}

	__pListener->OnIpcServerStarted(*this);

	g_main_loop_run(__pHandlerGMainLoop);

	__pListener->OnIpcServerStopped(*this);
}

result
IpcServer::Start(void)
{
	return E_SUCCESS;
}

String
IpcServer::GetName(void) const
{
	return __name;
}

result
IpcServer::Stop(void)
{
	result r = E_SUCCESS;
	int ret = 0;

	SysTryReturnResult(NID_IO, __pListener != null, E_SYSTEM, "Listener is null.");

	if (!__runOnCallerThread)
	{
		pthread_t self = pthread_self();

		if (__pHandlerGMainLoop)
		{
			g_main_loop_quit(__pHandlerGMainLoop);
		}

		if (__handlerThread != self)
		{
			ret = pthread_join(__handlerThread, null);
			SysTryLog(NID_IO, ret == 0, "Join an IPC thread returns an error");
		}
	}
	else
	{
		__pListener->OnIpcServerStopped(*this);
	}

	return r;
}

bool
IpcServer::Send(IPC::Message* msg)
{
	gsize remain = 0;
	gsize written = 0;
	char* pData = NULL;
	GError* pGError = NULL;


	pData = (char*) msg->data();
	remain = msg->size();

	if (msg->is_reply())
	{
		while (remain > 0)
		{
			pGError = NULL;
			g_io_channel_write_chars(__pCurrentChannel, (char*) pData, remain, &written, &pGError);

			remain -= written;
			pData += written;
		}

		g_io_channel_flush(__pCurrentChannel, &pGError);
	}
	else
	{
		// empty statement;
	}

	delete msg;

	return true;
}

result
IpcServer::SendResponse(int client, IPC::Message* pMessage)
{
	result r = E_SUCCESS;
	gsize remain = 0;
	gsize written = 0;
	char* pData = null;
	GError* pGError = null;
	_ClientInfo* pClientInfo = null;
	int ret = 0;

	SysTryReturn(NID_IO, client >= 0 && pMessage != null, false, E_INVALID_ARG,
				"[E_INVALID_ARG] pMessage(0x%x) is null or clinet(%d) < 0", pMessage,
				client);
	SysTryCatch(NID_IO, !pMessage->is_sync(), r = E_INVALID_ARG, E_INVALID_ARG, "[E_INVALID_ARG] Can't send sync. messagee.");

	pClientInfo = __clients[client];
	SysTryCatch(NID_IO, pClientInfo != null, r = E_INVALID_ARG, E_INVALID_ARG,
			   "[E_INVALID_ARG] client(%d) has not been registered.",
			   client);

	pData = (char*) pMessage->data();
	remain = pMessage->size();

	while (remain > 0)
	{
		pGError = NULL;
		ret = g_io_channel_write_chars(pClientInfo->pReverseChannel, (char*) pData, remain, &written, &pGError);
		if (ret != G_IO_STATUS_NORMAL)
		{
			SysLog(NID_IO, "Failed to send a response: %d", ret);
			SysTryCatch(NID_IO, ret != G_IO_STATUS_ERROR, , E_SYSTEM, "[E_SYSTEM] Error occurred during writing message to socket.");
		}

		remain -= written;
		pData += written;
	}

	g_io_channel_flush(pClientInfo->pReverseChannel, &pGError);

	delete pMessage;

	return E_SUCCESS;

CATCH:
	delete pMessage;
	return r;
}

result
IpcServer::SendResponse(int client, const IPC::Message& message)
{
	result r = E_SUCCESS;
	gsize remain = 0;
	gsize written = 0;
	char* pData = null;
	GError* pGError = null;
	_ClientInfo* pClientInfo = null;
	int ret = 0;

	SysTryReturn(NID_IO, client >= 0, false, E_INVALID_ARG, "[E_INVALID_ARG] clinet(%d) < 0", client);
	SysTryCatch(NID_IO, !message.is_sync(), r = E_INVALID_ARG, E_INVALID_ARG, "[E_INVALID_ARG] Can't send sync. messagee.");

	pClientInfo = __clients[client];
	SysTryCatch(NID_IO, pClientInfo != null, r = E_INVALID_ARG, E_INVALID_ARG,
			   "[E_INVALID_ARG] client(%d) has not been registered.",
			   client);

	pData = (char*) message.data();
	remain = message.size();

	while (remain > 0)
	{
		pGError = NULL;
		ret = g_io_channel_write_chars(pClientInfo->pReverseChannel, (char*) pData, remain, &written, &pGError);
		if (ret != G_IO_STATUS_NORMAL)
		{
			SysLog(NID_IO, "Failed to send a response: %d", ret);
			SysTryCatch(NID_IO, ret != G_IO_STATUS_ERROR, , E_SYSTEM, "[E_SYSTEM] Error occurred during writing message to socket.");
		}

		remain -= written;
		pData += written;
	}

	g_io_channel_flush(pClientInfo->pReverseChannel, &pGError);

	return E_SUCCESS;

CATCH:
	return r;
}
